using System;
using System.IO;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using XFileFormat = Microsoft.DirectX.Direct3D.XFileFormat;

namespace AMF
{
    public partial class ModelEx : Form
    {
        private RenderPanel renderWindow;
        private MenuStrip mainMenu;
        private ToolStripMenuItem fileMenu;
        private ToolStripMenuItem fileOpen;
        private ToolStripMenuItem fileSaveAs;
        private ToolStripMenuItem fileExit;
        private ComboBox texturesList;
        private TabOnlyButton loadTexture;
        private ToolStripMenuItem viewMenu;
        private ToolStripMenuItem viewResetPosition;
        private ToolStripMenuItem fileLoadTextures;
        private ToolStripMenuItem fileLoadTexturesFromFolder;
        private ToolStripMenuItem fileLoadTexturesFromFile;
        private ToolStripMenuItem fileLoadTexturesFromBigfile;
        private ToolStripMenuItem viewTextureFiltering;
        private ToolStripMenuItem viewWireframe;
        private ToolStripMenuItem viewShowTextures;
        private ToolStripMenuItem viewLightAndShade;
        private ToolStripMenuItem viewCullMode;
        private ToolStripMenuItem viewCullModeClockwise;
        private ToolStripMenuItem viewCullModeAntiClockwise;
        private ToolStripMenuItem viewCullModeNoCulling;
        private ToolStripMenuItem fileSaveAll;
        private ToolStripMenuItem viewResetAll;
        private MeshLoader GexMesh;

        public ModelEx(int width, int height, bool isWindowed, string name)
        {
            this.InitializeComponent();
            if (!isWindowed) this.FormBorderStyle = FormBorderStyle.None;
            this.Name = name;
            this.Text = name;
            this.GexMesh = new MeshLoader(renderWindow, isWindowed);
            this.mainMenu.BringToFront();
            this.renderWindow.PreviewKeyDown += new PreviewKeyDownEventHandler(KeyboardControls);
        }

        void KeyboardControls(object sender, PreviewKeyDownEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Add:
                case Keys.Oemplus:
                    GexMesh.ZoomIn();
                    break;
                case Keys.Subtract:
                case Keys.OemMinus:
                    GexMesh.ZoomOut();
                    break;
                case Keys.Up:
                case Keys.NumPad8:
                    GexMesh.RotateUp();
                    break;
                case Keys.Down:
                case Keys.NumPad2:
                    GexMesh.RotateDown();
                    break;
                case Keys.Left:
                case Keys.NumPad4:
                    GexMesh.RotateLeft();
                    break;
                case Keys.Right:
                case Keys.NumPad6:
                    GexMesh.RotateRight();
                    break;
            }
            return;
        }

        [STAThread]
        public static void Main()
        {
            ModelEx MainWindow = new ModelEx(800, 600, true, "SR1-ModelEx v2.0");
            MainWindow.Show();
            Application.Run(MainWindow);
            return;
        }

        private void InitializeComponent()
        {
            System.ComponentModel.ComponentResourceManager resources = new System.ComponentModel.ComponentResourceManager(typeof(ModelEx));
            this.mainMenu = new System.Windows.Forms.MenuStrip();
            this.fileMenu = new System.Windows.Forms.ToolStripMenuItem();
            this.fileOpen = new System.Windows.Forms.ToolStripMenuItem();
            this.fileSaveAs = new System.Windows.Forms.ToolStripMenuItem();
            this.fileSaveAll = new System.Windows.Forms.ToolStripMenuItem();
            this.fileLoadTextures = new System.Windows.Forms.ToolStripMenuItem();
            this.fileLoadTexturesFromFolder = new System.Windows.Forms.ToolStripMenuItem();
            this.fileLoadTexturesFromFile = new System.Windows.Forms.ToolStripMenuItem();
            this.fileLoadTexturesFromBigfile = new System.Windows.Forms.ToolStripMenuItem();
            this.fileExit = new System.Windows.Forms.ToolStripMenuItem();
            this.viewMenu = new System.Windows.Forms.ToolStripMenuItem();
            this.viewResetPosition = new System.Windows.Forms.ToolStripMenuItem();
            this.viewCullMode = new System.Windows.Forms.ToolStripMenuItem();
            this.viewCullModeClockwise = new System.Windows.Forms.ToolStripMenuItem();
            this.viewCullModeAntiClockwise = new System.Windows.Forms.ToolStripMenuItem();
            this.viewCullModeNoCulling = new System.Windows.Forms.ToolStripMenuItem();
            this.viewLightAndShade = new System.Windows.Forms.ToolStripMenuItem();
            this.viewShowTextures = new System.Windows.Forms.ToolStripMenuItem();
            this.viewTextureFiltering = new System.Windows.Forms.ToolStripMenuItem();
            this.viewWireframe = new System.Windows.Forms.ToolStripMenuItem();
            this.texturesList = new System.Windows.Forms.ComboBox();
            this.loadTexture = new AMF.TabOnlyButton();
            this.renderWindow = new AMF.RenderPanel();
            this.viewResetAll = new System.Windows.Forms.ToolStripMenuItem();
            this.mainMenu.SuspendLayout();
            this.SuspendLayout();
            // 
            // mainMenu
            // 
            this.mainMenu.Items.AddRange(new System.Windows.Forms.ToolStripItem[] {
            this.fileMenu,
            this.viewMenu});
            this.mainMenu.Location = new System.Drawing.Point(0, 0);
            this.mainMenu.Name = "mainMenu";
            this.mainMenu.Size = new System.Drawing.Size(594, 24);
            this.mainMenu.TabIndex = 1;
            // 
            // fileMenu
            // 
            this.fileMenu.DropDownItems.AddRange(new System.Windows.Forms.ToolStripItem[] {
            this.fileOpen,
            this.fileSaveAs,
            this.fileSaveAll,
            this.fileLoadTextures,
            this.fileExit});
            this.fileMenu.Name = "fileMenu";
            this.fileMenu.Size = new System.Drawing.Size(35, 20);
            this.fileMenu.Text = "File";
            // 
            // fileOpen
            // 
            this.fileOpen.Name = "fileOpen";
            this.fileOpen.Size = new System.Drawing.Size(154, 22);
            this.fileOpen.Text = "Open...";
            this.fileOpen.Click += new System.EventHandler(this.fileOpen_Click);
            // 
            // fileSaveAs
            // 
            this.fileSaveAs.Enabled = false;
            this.fileSaveAs.Name = "fileSaveAs";
            this.fileSaveAs.Size = new System.Drawing.Size(154, 22);
            this.fileSaveAs.Text = "Save As...";
            this.fileSaveAs.Click += new System.EventHandler(this.fileSaveAs_Click);
            // 
            // fileSaveAll
            // 
            this.fileSaveAll.Enabled = false;
            this.fileSaveAll.Name = "fileSaveAll";
            this.fileSaveAll.Size = new System.Drawing.Size(154, 22);
            this.fileSaveAll.Text = "Save All...";
            this.fileSaveAll.Click += new System.EventHandler(this.fileSaveAll_Click);
            // 
            // fileLoadTextures
            // 
            this.fileLoadTextures.DropDownItems.AddRange(new System.Windows.Forms.ToolStripItem[] {
            this.fileLoadTexturesFromFolder,
            this.fileLoadTexturesFromFile,
            this.fileLoadTexturesFromBigfile});
            this.fileLoadTextures.Enabled = false;
            this.fileLoadTextures.Name = "fileLoadTextures";
            this.fileLoadTextures.Size = new System.Drawing.Size(154, 22);
            this.fileLoadTextures.Text = "Load Textures";
            // 
            // fileLoadTexturesFromFolder
            // 
            this.fileLoadTexturesFromFolder.Name = "fileLoadTexturesFromFolder";
            this.fileLoadTexturesFromFolder.Size = new System.Drawing.Size(186, 22);
            this.fileLoadTexturesFromFolder.Text = "From Folder...";
            this.fileLoadTexturesFromFolder.Click += new System.EventHandler(this.fileLoadTexturesFromFolder_Click);
            // 
            // fileLoadTexturesFromFile
            // 
            this.fileLoadTexturesFromFile.Enabled = false;
            this.fileLoadTexturesFromFile.Name = "fileLoadTexturesFromFile";
            this.fileLoadTexturesFromFile.Size = new System.Drawing.Size(186, 22);
            this.fileLoadTexturesFromFile.Text = "From Textures File...";
            // 
            // fileLoadTexturesFromBigfile
            // 
            this.fileLoadTexturesFromBigfile.Enabled = false;
            this.fileLoadTexturesFromBigfile.Name = "fileLoadTexturesFromBigfile";
            this.fileLoadTexturesFromBigfile.Size = new System.Drawing.Size(186, 22);
            this.fileLoadTexturesFromBigfile.Text = "From Bigfile...";
            // 
            // fileExit
            // 
            this.fileExit.Name = "fileExit";
            this.fileExit.Size = new System.Drawing.Size(154, 22);
            this.fileExit.Text = "Exit";
            this.fileExit.Click += new System.EventHandler(this.fileExit_Click);
            // 
            // viewMenu
            // 
            this.viewMenu.DropDownItems.AddRange(new System.Windows.Forms.ToolStripItem[] {
            this.viewResetAll,
            this.viewResetPosition,
            this.viewCullMode,
            this.viewLightAndShade,
            this.viewShowTextures,
            this.viewTextureFiltering,
            this.viewWireframe});
            this.viewMenu.Name = "viewMenu";
            this.viewMenu.Size = new System.Drawing.Size(41, 20);
            this.viewMenu.Text = "View";
            // 
            // viewResetPosition
            // 
            this.viewResetPosition.Name = "viewResetPosition";
            this.viewResetPosition.Size = new System.Drawing.Size(164, 22);
            this.viewResetPosition.Text = "Reset Position";
            this.viewResetPosition.Click += new System.EventHandler(this.viewResetPosition_Click);
            // 
            // viewCullMode
            // 
            this.viewCullMode.DropDownItems.AddRange(new System.Windows.Forms.ToolStripItem[] {
            this.viewCullModeClockwise,
            this.viewCullModeAntiClockwise,
            this.viewCullModeNoCulling});
            this.viewCullMode.Name = "viewCullMode";
            this.viewCullMode.Size = new System.Drawing.Size(164, 22);
            this.viewCullMode.Text = "Cull Mode";
            // 
            // viewCullModeClockwise
            // 
            this.viewCullModeClockwise.Name = "viewCullModeClockwise";
            this.viewCullModeClockwise.Size = new System.Drawing.Size(154, 22);
            this.viewCullModeClockwise.Text = "Clockwise";
            this.viewCullModeClockwise.Click += new System.EventHandler(this.viewCullModeClockwise_Click);
            // 
            // viewCullModeAntiClockwise
            // 
            this.viewCullModeAntiClockwise.Name = "viewCullModeAntiClockwise";
            this.viewCullModeAntiClockwise.Size = new System.Drawing.Size(154, 22);
            this.viewCullModeAntiClockwise.Text = "Anti-Clockwise";
            this.viewCullModeAntiClockwise.Click += new System.EventHandler(this.viewCullModeAntiClockwise_Click);
            // 
            // viewCullModeNoCulling
            // 
            this.viewCullModeNoCulling.Checked = true;
            this.viewCullModeNoCulling.CheckState = System.Windows.Forms.CheckState.Checked;
            this.viewCullModeNoCulling.Name = "viewCullModeNoCulling";
            this.viewCullModeNoCulling.Size = new System.Drawing.Size(154, 22);
            this.viewCullModeNoCulling.Text = "No Culling";
            this.viewCullModeNoCulling.Click += new System.EventHandler(this.viewCullModeNoCulling_Click);
            // 
            // viewLightAndShade
            // 
            this.viewLightAndShade.Checked = true;
            this.viewLightAndShade.CheckState = System.Windows.Forms.CheckState.Checked;
            this.viewLightAndShade.Name = "viewLightAndShade";
            this.viewLightAndShade.Size = new System.Drawing.Size(164, 22);
            this.viewLightAndShade.Text = "Lighting/Shading";
            this.viewLightAndShade.Click += new System.EventHandler(this.viewLightAndShade_Click);
            // 
            // viewShowTextures
            // 
            this.viewShowTextures.Checked = true;
            this.viewShowTextures.CheckState = System.Windows.Forms.CheckState.Checked;
            this.viewShowTextures.Name = "viewShowTextures";
            this.viewShowTextures.Size = new System.Drawing.Size(164, 22);
            this.viewShowTextures.Text = "Show Textures";
            this.viewShowTextures.Click += new System.EventHandler(this.viewShowTextures_Click);
            // 
            // viewTextureFiltering
            // 
            this.viewTextureFiltering.Checked = true;
            this.viewTextureFiltering.CheckState = System.Windows.Forms.CheckState.Checked;
            this.viewTextureFiltering.Name = "viewTextureFiltering";
            this.viewTextureFiltering.Size = new System.Drawing.Size(164, 22);
            this.viewTextureFiltering.Text = "Texture Filtering";
            this.viewTextureFiltering.Click += new System.EventHandler(this.viewTextureFiltering_Click);
            // 
            // viewWireframe
            // 
            this.viewWireframe.Name = "viewWireframe";
            this.viewWireframe.Size = new System.Drawing.Size(164, 22);
            this.viewWireframe.Text = "Wireframe";
            this.viewWireframe.Click += new System.EventHandler(this.viewWireframe_Click);
            // 
            // texturesList
            // 
            this.texturesList.DropDownStyle = System.Windows.Forms.ComboBoxStyle.DropDownList;
            this.texturesList.FormattingEnabled = true;
            this.texturesList.Location = new System.Drawing.Point(50, 40);
            this.texturesList.Name = "texturesList";
            this.texturesList.Size = new System.Drawing.Size(425, 21);
            this.texturesList.TabIndex = 2;
            // 
            // loadTexture
            // 
            this.loadTexture.Enabled = false;
            this.loadTexture.Location = new System.Drawing.Point(480, 40);
            this.loadTexture.Name = "loadTexture";
            this.loadTexture.Size = new System.Drawing.Size(70, 21);
            this.loadTexture.TabIndex = 3;
            this.loadTexture.Text = "Load...";
            this.loadTexture.UseVisualStyleBackColor = true;
            this.loadTexture.Click += new System.EventHandler(this.fileLoadTexture_Click);
            // 
            // renderWindow
            // 
            this.renderWindow.BackColor = System.Drawing.Color.Gray;
            this.renderWindow.BorderStyle = System.Windows.Forms.BorderStyle.Fixed3D;
            this.renderWindow.Location = new System.Drawing.Point(50, 70);
            this.renderWindow.Name = "renderWindow";
            this.renderWindow.Size = new System.Drawing.Size(500, 500);
            this.renderWindow.TabIndex = 0;
            this.renderWindow.TabStop = true;
            // 
            // viewResetAll
            // 
            this.viewResetAll.Name = "viewResetAll";
            this.viewResetAll.Size = new System.Drawing.Size(164, 22);
            this.viewResetAll.Text = "Reset All";
            this.viewResetAll.Click += new System.EventHandler(this.viewResetAll_Click);
            // 
            // ModelEx
            // 
            this.ClientSize = new System.Drawing.Size(594, 625);
            this.Controls.Add(this.loadTexture);
            this.Controls.Add(this.texturesList);
            this.Controls.Add(this.renderWindow);
            this.Controls.Add(this.mainMenu);
            this.FormBorderStyle = System.Windows.Forms.FormBorderStyle.FixedDialog;
            this.Icon = ((System.Drawing.Icon)(resources.GetObject("$this.Icon")));
            this.MainMenuStrip = this.mainMenu;
            this.MaximizeBox = false;
            this.Name = "ModelEx";
            this.Text = "SR1-ModelEx v2.0";
            this.mainMenu.ResumeLayout(false);
            this.mainMenu.PerformLayout();
            this.ResumeLayout(false);
            this.PerformLayout();

        }

        private void UpdateTextureList()
        {
            texturesList.Items.Clear();
            string[] textures = GexMesh.TextureNames;
            if (textures.Length == 0) return;
            for (int i = 0; i < textures.Length; i++)
            {
                texturesList.Items.Add(textures[i]);
            }
            texturesList.SelectedIndex = 0;
            return;
        }

        private void fileOpen_Click(object sender, EventArgs e)
        {
            bool succeded = true;
            GexMesh.pauseRendering = true;
            OpenFileDialog OpenDlg = new OpenFileDialog();
            OpenDlg.CheckFileExists = true;
            OpenDlg.CheckPathExists = true;
            OpenDlg.Filter = "Soul Reaver Mesh Files|*.SRObj;*.drm;*.pcm|" +
                                "Soul Reaver DRM Files (*.drm)|*.drm|" +
                                "Soul Reaver PCM Files (*.pcm)|*.pcm|" +
                                "DirectX Mesh Files (*.x)|*.x|" +
                                "All Files (*.*)|*.*";
            OpenDlg.DefaultExt = "drm";
            if (OpenDlg.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    if (OpenDlg.FilterIndex < 3) GexMesh.LoadGexMesh(OpenDlg.FileName);
                    else GexMesh.LoadDXMesh(OpenDlg.FileName);
                }
                catch (Exception error)
                {
                    MessageBox.Show(
                        error.Message,
                        "Could not load the file",
                        MessageBoxButtons.OK,
                        MessageBoxIcon.Information
                    );
                    // Why does it need this?
                    renderWindow.Invalidate();
                    succeded = false;
                }
                UpdateTextureList();
            }
            GexMesh.pauseRendering = false;
            if (succeded)
            {
                loadTexture.Enabled = true;
                fileLoadTextures.Enabled = true;
                fileSaveAs.Enabled = true;
                fileSaveAll.Enabled = true;
                viewResetAll_Click(sender, e);
            }
            renderWindow.Focus();
            return;
        }

        private void fileLoadTexture_Click(object sender, EventArgs e)
        {
            GexMesh.pauseRendering = true;
            OpenFileDialog OpenDlg = new OpenFileDialog();
            OpenDlg.CheckFileExists = true;
            OpenDlg.CheckPathExists = true;
            OpenDlg.Filter = "PNG (*.PNG)|*.PNG";
            OpenDlg.DefaultExt = "PNG";
            if (OpenDlg.ShowDialog() == DialogResult.OK)
            {
                GexMesh.LoadTexture(OpenDlg.FileName, texturesList.SelectedIndex);
                UpdateTextureList();
            }
            GexMesh.pauseRendering = false;
            renderWindow.Focus();
            return;
        }

        private void fileLoadTexturesFromFolder_Click(object sender, EventArgs e)
        {
            GexMesh.pauseRendering = true;
            FolderBrowserDialog FolderDlg = new FolderBrowserDialog();
            FolderDlg.Description = "Load all required textures from the current folder";
            FolderDlg.ShowNewFolderButton = false;
            FolderDlg.SelectedPath = Directory.GetCurrentDirectory();
            if (FolderDlg.ShowDialog() == DialogResult.OK)
            {
                GexMesh.LoadTexturesFromFolder(FolderDlg.SelectedPath);
            }
            GexMesh.pauseRendering = false;
            renderWindow.Focus();
            return;
        }

        private void fileExit_Click(object sender, EventArgs e)
        {
            this.Close();
            return;
        }

        private void fileSaveAs_Click(object sender, EventArgs e)
        {
            GexMesh.pauseRendering = true;
            SaveFileDialog SaveDlg = new SaveFileDialog();
            SaveDlg.Filter = "DirectX Mesh Files (Text)(*.x)|*.x|" +
                                "DirectX Mesh Files (Binary)(*.x)|*.x|" +
                                "DirectX Mesh Files (Compressed)(*.x)|*.x";
            SaveDlg.DefaultExt = "x";
            SaveDlg.FilterIndex = 3;
            SaveDlg.FileName = GexMesh.name + ".x";
            if (sender != fileSaveAs)
            {
                SaveDlg.CheckFileExists = false;
                SaveDlg.OverwritePrompt = false;
                SaveDlg.FileOk += new CancelEventHandler(SaveDlg_FileOk);
            }
            if (SaveDlg.ShowDialog() == DialogResult.OK)
            {
                XFileFormat format = XFileFormat.Text;
                switch (SaveDlg.FilterIndex)
                {
                    case 1:
                        format = XFileFormat.Text;
                        break;
                    case 2:
                        format = XFileFormat.Binary;
                        break;
                    case 3:
                        format = XFileFormat.Compressed;
                        break;
                }
                GexMesh.SaveDXMesh(
                    SaveDlg.FileName,
                    format,
                    sender != fileSaveAs,
                    false
                );
            }
            GexMesh.pauseRendering = false;
            renderWindow.Focus();
            renderWindow.Invalidate();
            return;
        }

        private void fileSaveAll_Click(object sender, EventArgs e)
        {
            fileSaveAs_Click(sender, e);
            return;
        }

        private void SaveDlg_FileOk(object sender, CancelEventArgs e)
        {
            SaveFileDialog SaveDlg = (SaveFileDialog)sender;
            String[] existingFiles = GexMesh.FindExistingFiles(SaveDlg.FileName);
            for (int f = 0; f < existingFiles.Length; f++)
            {
                if (existingFiles[f] != null)
                {
                    DialogResult overwrite = MessageBox.Show(
                        "File " + existingFiles[f] + " already exists.\r\n" +
                        "Do you want to replace it?",
                        "Save As",
                        MessageBoxButtons.YesNo,
                        MessageBoxIcon.Exclamation
                    );
                    e.Cancel = (overwrite == DialogResult.No);
                }
                if (e.Cancel)
                {
                    MessageBox.Show(
                        "No files were overwritten",
                        "Save Cancelled",
                        MessageBoxButtons.OK,
                        MessageBoxIcon.Information
                       );
                    return;
                }
            }
        }

        private void viewResetAll_Click(object sender, EventArgs e)
        {
            if (sender == viewResetAll) GexMesh.ResetViewModes();
            viewCullModeClockwise.Checked = false;
            viewCullModeAntiClockwise.Checked = false;
            viewCullModeNoCulling.Checked = true;
            viewLightAndShade.Checked = true;
            viewShowTextures.Checked = true;
            viewWireframe.Checked = false;
            viewTextureFiltering.Checked = true;
        }

        private void viewResetPosition_Click(object sender, EventArgs e)
        {
            GexMesh.ResetTransforms();
            return;
        }

        private void viewTextureFiltering_Click(object sender, EventArgs e)
        {
            viewTextureFiltering.Checked ^= true;
            GexMesh.FilterTextures = viewTextureFiltering.Checked;
        }

        private void viewWireframe_Click(object sender, EventArgs e)
        {
            viewWireframe.Checked ^= true;
            GexMesh.WireFrameMode = viewWireframe.Checked;
        }

        private void viewShowTextures_Click(object sender, EventArgs e)
        {
            viewShowTextures.Checked ^= true;
            GexMesh.UseTextures = viewShowTextures.Checked;
        }

        private void viewLightAndShade_Click(object sender, EventArgs e)
        {
            viewLightAndShade.Checked ^= true;
            GexMesh.UseLighting = viewLightAndShade.Checked;
        }

        private void viewCullModeClockwise_Click(object sender, EventArgs e)
        {
            GexMesh.SetCullMode(MeshLoader.CullMode.Clockwise);
            viewCullModeClockwise.Checked = true;
            viewCullModeAntiClockwise.Checked = false;
            viewCullModeNoCulling.Checked = false;
        }

        private void viewCullModeAntiClockwise_Click(object sender, EventArgs e)
        {
            GexMesh.SetCullMode(MeshLoader.CullMode.CounterClockwise);
            viewCullModeClockwise.Checked = false;
            viewCullModeAntiClockwise.Checked = true;
            viewCullModeNoCulling.Checked = false;
        }

        private void viewCullModeNoCulling_Click(object sender, EventArgs e)
        {
            GexMesh.SetCullMode(MeshLoader.CullMode.None);
            viewCullModeClockwise.Checked = false;
            viewCullModeAntiClockwise.Checked = false;
            viewCullModeNoCulling.Checked = true;
        }
    }

    public class RenderPanel : Panel
    {
        private bool okToLeave = true;
        public RenderPanel()
        {
            this.SetStyle(ControlStyles.Selectable, true);
            this.SetStyle(ControlStyles.AllPaintingInWmPaint | ControlStyles.Opaque, true);
        }
        protected override void OnClick(EventArgs e)
        {
            Focus();
            base.OnClick(e);
            return;
        }
        // Try IsInputKey(Keys keyData) and return true if not tab
        protected override void OnLostFocus(EventArgs e)
        {
            if (!okToLeave)
            {
                Focus();
                okToLeave = true;
            }
            base.OnLostFocus(e);
            return;
        }
        protected override void OnPreviewKeyDown(PreviewKeyDownEventArgs e)
        {
            if (e.KeyCode != Keys.Tab)
            {
                okToLeave = false;
            }
            base.OnPreviewKeyDown(e);
            return;
        }
    }

    public class TabOnlyButton : Button
    {
        private bool okToLeave = true;

        protected override void OnLostFocus(EventArgs e)
        {
            if (!okToLeave)
            {
                Focus();
                okToLeave = true;
            }
            base.OnLostFocus(e);
            return;
        }

        protected override void OnPreviewKeyDown(PreviewKeyDownEventArgs e)
        {
            if (e.KeyCode != Keys.Tab)
            {
                okToLeave = false;
            }
            base.OnPreviewKeyDown(e);
            return;
        }
    }
}